import pandas as pd
from csv_pm_llm_parsing import constants, util
import traceback
import time
from typing import Optional, Union, Dict


def detect_timest_format(df: pd.DataFrame, timest_column: str = "time:timestamp", max_head_n: int = 10,
                         max_retry: int = constants.MAX_RETRY, openai_api_url: Optional[str] = None,
                         openai_api_key: Optional[str] = None,
                         openai_model: Optional[str] = None, return_timest_format: bool = False, debug: bool = False) -> Union[pd.DataFrame, Dict[str, str]]:
    """
    Automatically detects the format of the timestamp in the specified column using LLMs.
    The Pandas dataframe's column is then parsed using the given format.

    Parameters
    ---------------
    df
        Pandas dataframe
    timest_column
        Column to which we aim to apply the automatic timestamp format detection
    max_head_n
        Number of top values that should be provided to the LLM
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_timest_format
        (bool) Returns the timestamp format (instead of the transformed dataframe)
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe (with the timestamp column parsed)
    """
    df[timest_column] = df[timest_column].astype(str).apply(lambda x: x.strip())
    head_values = list(df[timest_column].head(min(len(df), max_head_n)))

    prompt = "Given the following list of timestamps: " + str(head_values)
    prompt += "\n\nCan you propose the timestamp format that I should use as parameter for Pandas timestamp parsing?"
    prompt += "\n\nPlease include the proposed format in a JSON, like: {\"format\": PROPOSEDFORMAT}"

    parsed = False
    for i in range(max_retry):
        try:
            proposed_format = util.get_json(
                util.openai_inquiry(prompt.encode('utf-8', errors='ignore').decode('utf-8'),
                                    openai_api_url=openai_api_url, openai_api_key=openai_api_key,
                                    openai_model=openai_model))["format"]

            if debug:
                print(proposed_format)
            if return_timest_format:
                return proposed_format

            df[timest_column] = pd.to_datetime(df[timest_column], format=proposed_format)
            parsed = True
            break
        except Exception as e:
            traceback.print_exc()
            prompt += "\n\nI am getting the following error: " + str(e)
            time.sleep(constants.SLEEP_TIME)

    if not parsed:
        raise Exception("failed parsing of the timestamp column: " + timest_column)

    return df



from csv_pm_llm_parsing import constants, util
import pandas as pd
from typing import Optional, Union, Dict
import traceback
import time


def detect_sep_and_load(file_path: str, input_encoding: str = "utf-8", read_bytes: int = 2048,
                        max_retry: int = constants.MAX_RETRY, openai_api_url: Optional[str] = None,
                        openai_api_key: Optional[str] = None,
                        openai_model: Optional[str] = None, return_detected_sep: bool = False, debug: bool = False) -> Union[pd.DataFrame, Dict[str, str]]:
    """
    Detects the separator and quotechar in the provided file using LLMs.

    Parameters
    ----------------
    file_path
        Path to the file
    input_encoding
        Encoding of the file (default: utf-8)
    read_bytes
        Number of bytes that should be initially considered
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_detected_sep
        (bool) Returns the detected separator and quotechar, instead of the Pandas dataframe
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe
    """
    F = open(file_path, 'r', encoding=input_encoding)
    first_characters = F.read(read_bytes)
    F.close()

    prompt = "Given the following contents of a CSV file:\n\n"
    prompt += first_characters
    prompt += "\n\n"
    prompt += "Could you identify the separator and quote characters for the Pandas read_csv method?\n"
    prompt += "Please produce a JSON including two keys, 'sep' and 'quotechar'."
    dataframe = None
    for i in range(max_retry):
        try:
            format = util.get_json(util.openai_inquiry(prompt.encode('utf-8', errors='ignore').decode('utf-8'),
                                                       openai_api_url=openai_api_url, openai_api_key=openai_api_key,
                                                       openai_model=openai_model))
            format['sep']
            format['quotechar']

            if debug:
                print(format)
            if return_detected_sep:
                return format

            dataframe = pd.read_csv(file_path, encoding=input_encoding, sep=format['sep'],
                                    quotechar=format['quotechar'])
            break
        except Exception as e:
            traceback.print_exc()
            prompt += "\n\nI am getting the following error: " + str(e)
            time.sleep(constants.SLEEP_TIME)

    if dataframe is None:
        raise Exception("failed detection of separator and quotechar.")

    return dataframe



import json, requests
from csv_pm_llm_parsing import constants
from typing import Optional, Dict, Any
import traceback


def openai_inquiry(prompt: str, openai_api_url: Optional[str] = None, openai_api_key: Optional[str] = None,
                   openai_model: Optional[str] = None) -> str:
    """
    Sends the prompt to the OpenAI's APIs, obtaining the response

    Parameters
    ------------------
    prompt
        Prompt
    openai_api_url
        OpenAI API url
    openai_api_key
        OpenAI API key
    openai_model
        OpenAI model

    Returns
    -----------------
    response
        Response from the APIs
    """
    if openai_api_url is None:
        openai_api_url = constants.OPENAI_API_URL

    if openai_api_key is None:
        openai_api_key = constants.OPENAI_API_KEY

    if openai_model is None:
        openai_model = constants.OPENAI_MODEL

    if openai_api_key is None or not openai_api_key:
        raise Exception(
            "Please provide the 'openai_api_key' parameter to the method or set the OPENAI_API_KEY environment variable!")

    if openai_model is None or not openai_model:
        raise Exception(
            "Please provide the 'openai_model' parameter to the method or set the OPENAI_MODEL environment variable!")

    if openai_api_url.endswith("/"):
        openai_api_url = openai_api_url[:-1]

    complete_url = openai_api_url + "/chat/completions"

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {openai_api_key}"
    }

    payload = {
        "model": openai_model,
        "messages": [{"role": "user", "content": prompt}]
    }

    response = requests.post(complete_url, headers=headers, json=payload).json()

    if "error" in response:
        # raise an exception when the request fails, with the provided message
        raise Exception(response["error"]["message"])

    return response["choices"][0]["message"]["content"]


def get_json(response: str) -> Dict[str, Any]:
    """
    Gets the last JSON dictionary contained in the response from the APIs

    Parameters
    ---------------
    response
        Response from the APIs

    Returns
    ---------------
    json_dict
        Last JSON dict
    """
    try:
        json_str = "{" + response.split("{")[-1].split("}")[0].strip() + "}"
        return json.loads(json_str)
    except:
        traceback.print_exc()
        return None



import pandas as pd
from csv_pm_llm_parsing import constants, util
from typing import Optional, Dict, Any, Union
import io, traceback, time


def detect_caseid_activity_timestamp(df: pd.DataFrame, max_retry: int = constants.MAX_RETRY,
                                     openai_api_url: Optional[str] = None,
                                     openai_api_key: Optional[str] = None,
                                     openai_model: Optional[str] = None, return_suggestions: bool = False, debug: bool = False) -> Union[
    pd.DataFrame, Dict[str, str]]:
    """
    Detects automatically the columns to use as case identifier, activity, and timestamp in the provided dataframe.

    Parameters
    -----------------
    df
        Pandas dataframe
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_suggestions
        (bool) Return the suggestion (dictionary) instead of the dataframe
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe with standard column names (i.e., in pm4py, 'case:concept:name' for the case ID, 'concept:name' for the activity, and 'time:timestamp' for the timestamp).
    """
    buf = io.StringIO()
    df.info(buf=buf)
    s = buf.getvalue()
    prompt = "Given the dataframe with the following columns:\n\n"
    prompt += s
    prompt += "\n\nCan you suggest some columns for the case identifier, activity, and completion timestamp?\n"
    prompt += "Please produce a JSON containing as keys: 'caseid', 'activity', 'timestamp'\n"
    prompt += "Each key should be associated with the name of the column."

    suggested = False
    for i in range(max_retry):
        try:
            suggestions = util.get_json(util.openai_inquiry(prompt.encode('utf-8', errors='ignore').decode('utf-8'),
                                                            openai_api_url=openai_api_url,
                                                            openai_api_key=openai_api_key,
                                                            openai_model=openai_model))
            df[suggestions['caseid']]
            df[suggestions['activity']]
            df[suggestions['timestamp']]
            suggested = True

            if debug:
                print(suggestions)
            if return_suggestions:
                return suggestions

            df['case:concept:name'] = df[suggestions['caseid']]
            df['concept:name'] = df[suggestions['activity']]
            df['time:timestamp'] = df[suggestions['timestamp']]
        except Exception as e:
            traceback.print_exc()
            prompt += "\n\nI am getting the following error: " + str(e)
            time.sleep(constants.SLEEP_TIME)

    if not suggested:
        raise Exception("failed process mining column detection")

    return df



import pandas as pd
from csv_pm_llm_parsing import constants, meta
from typing import Optional, Union, Dict
import sys, traceback


def detect_timest_format(df: pd.DataFrame, timest_column: str = "time:timestamp", max_head_n: int = 10,
                         max_retry: int = constants.MAX_RETRY, openai_api_url: Optional[str] = None,
                         openai_api_key: Optional[str] = None,
                         openai_model: Optional[str] = None, return_timest_format: bool = False, debug: bool = False) -> Union[
    pd.DataFrame, Dict[str, str]]:
    """
    Automatically detects the format of the timestamp in the specified column using LLMs.
    The Pandas dataframe's column is then parsed using the given format.

    Parameters
    ---------------
    df
        Pandas dataframe
    timest_column
        Column to which we aim to apply the automatic timestamp format detection
    max_head_n
        Number of top values that should be provided to the LLM
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_timest_format
        (bool) Returns the timestamp format (instead of the transformed dataframe)
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe (with the timestamp column parsed)
    """
    from csv_pm_llm_parsing import timest_parser
    return timest_parser.detect_timest_format(df, timest_column=timest_column, max_head_n=max_head_n,
                                              max_retry=max_retry, openai_api_url=openai_api_url,
                                              openai_api_key=openai_api_key, openai_model=openai_model,
                                              return_timest_format=return_timest_format, debug=debug)


def detect_sep_and_quote(file_path: str, input_encoding: str = "utf-8", read_bytes: int = 2048,
                         max_retry: int = constants.MAX_RETRY, openai_api_url: Optional[str] = None,
                         openai_api_key: Optional[str] = None,
                         openai_model: Optional[str] = None, return_detected_sep: bool = False, debug: bool = False) -> Union[
    pd.DataFrame, Dict[str, str]]:
    """
    Detects the separator and quotechar in the provided file using LLMs.

    Parameters
    ----------------
    file_path
        Path to the file
    input_encoding
        Encoding of the file (default: utf-8)
    read_bytes
        Number of bytes that should be initially considered
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_detected_sep
        (bool) Returns the detected separator and quotechar, instead of the Pandas dataframe
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe
    """
    from csv_pm_llm_parsing import sep_detection
    return sep_detection.detect_sep_and_load(file_path, input_encoding=input_encoding, read_bytes=read_bytes,
                                             max_retry=max_retry, openai_api_url=openai_api_url,
                                             openai_api_key=openai_api_key, openai_model=openai_model,
                                             return_detected_sep=return_detected_sep, debug=debug)


def detect_caseid_activity_timestamp(df: pd.DataFrame, max_retry: int = constants.MAX_RETRY,
                                     openai_api_url: Optional[str] = None,
                                     openai_api_key: Optional[str] = None,
                                     openai_model: Optional[str] = None, return_suggestions: bool = False, debug: bool = False) -> Union[
    pd.DataFrame, Dict[str, str]]:
    """
    Detects automatically the columns to use as case identifier, activity, and timestamp in the provided dataframe.

    Parameters
    -----------------
    df
        Pandas dataframe
    max_retry
        Maximum number of retries upon failure
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    return_suggestions
        (bool) Return the suggestion (dictionary) instead of the dataframe
    debug
        (bool) Prints additional debug information

    Returns
    ----------------
    df
        Pandas dataframe with standard column names (i.e., in pm4py, 'case:concept:name' for the case ID, 'concept:name' for the activity, and 'time:timestamp' for the timestamp).
    """
    from csv_pm_llm_parsing import pm_columns_detection
    return pm_columns_detection.detect_caseid_activity_timestamp(df, max_retry=max_retry, openai_api_url=openai_api_url,
                                                                 openai_api_key=openai_api_key,
                                                                 openai_model=openai_model,
                                                                 return_suggestions=return_suggestions,
                                                                 debug=debug)


def __parse_bytes(file_path: str, n_bytes: int) -> str:
    """
    Parses the specific number of bytes from the file to retrieve the encoding
    (using the 'chardet' package)

    Parameters
    --------------
    file_path
        Path to the CSV file
    n_bytes
        Number of bytes to decide the encoding

    Returns
    --------------
    encoding
        Encoding
    """
    import chardet

    F = open(file_path, "rb")
    byte_content = F.read(n_bytes)
    F.close()

    result = chardet.detect(byte_content)
    encoding = result['encoding']

    F = open(file_path, "r", encoding=encoding)
    F.read()
    F.close()

    return encoding


def full_parse_csv_for_pm(file_path: str, openai_api_url: Optional[str] = None,
                          openai_api_key: Optional[str] = None,
                          openai_model: Optional[str] = None, debug: bool = False) -> pd.DataFrame:
    """
    Starting from the specified path, reads the CSV in a process-mining-ready format.

    Parameters
    ---------------
    file_path
        Path to the CSV file
    openai_api_url
        API URL (like https://api.openai.com/v1 or http://127.0.0.1:11434/v1 )
    openai_api_key
        API key
    openai_model
        OpenAI model
    debug
        (bool) Prints additional debug information

    Returns
    ---------------
    dataframe
        Pandas dataframe
    """
    try:
        encoding = __parse_bytes(file_path, 2048)
    except:
        encoding = __parse_bytes(file_path, sys.maxsize)

    if debug:
        print(encoding)

    for i in range(constants.MAX_RETRY):
        try:
            dataframe = detect_sep_and_quote(file_path, input_encoding=encoding, openai_api_url=openai_api_url,
                                             openai_api_key=openai_api_key, openai_model=openai_model, debug=debug, max_retry=1)

            dataframe = detect_caseid_activity_timestamp(dataframe, openai_api_url=openai_api_url,
                                                         openai_api_key=openai_api_key, openai_model=openai_model, debug=debug, max_retry=1)
            dataframe = dataframe.dropna(subset=["case:concept:name", "concept:name", "time:timestamp"])
            dataframe["case:concept:name"] = dataframe["case:concept:name"].astype(str)
            dataframe["concept:name"] = dataframe["concept:name"].astype(str)

            dataframe = detect_timest_format(dataframe, "time:timestamp", openai_api_url=openai_api_url,
                                             openai_api_key=openai_api_key, openai_model=openai_model, debug=debug, max_retry=2)

            dataframe["@@index"] = dataframe.index
            dataframe.sort_values(["case:concept:name", "time:timestamp", "@@index"])

            break
        except:
            traceback.print_exc()

    return dataframe



__name__ = 'csv-pm-llm-parsing'
VERSION = '0.1'
__version__ = VERSION
__doc__ = 'LLM-based CSV parsing for Process Mining purposes'
__author__ = 'Alessandro Berti, PADS @ RWTH'
__author_email__ = 'a.berti@pads.rwth-aachen.de'
__maintainer__ = 'Alessandro Berti, PADS @ RWTH'
__maintainer_email__ = "a.berti@pads.rwth-aachen.de"



import os

SLEEP_TIME = 3
MAX_RETRY = 5
OPENAI_API_URL = os.environ.get("OPENAI_API_URL", "https://api.openai.com/v1")
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)
OPENAI_MODEL = os.environ.get("OPENAI_MODEL", None)
